@@ -1,8 +1,8 @@
 ;;; yasnippet-tests.el --- some yasnippet tests  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2012, 2013, 2014, 2015, 2017  Free Software Foundation, Inc.
+;; Copyright (C) 2012-2024  Free Software Foundation, Inc.
 
-;; Author: Joo Tvora <joaot@siscog.pt>
+;; Author: João Távora <joaot@siscog.pt>
 ;; Keywords: emulations, convenience
 
 ;; This program is free software; you can redistribute it and/or modify
@@ -121,7 +121,8 @@ This lets `yas--maybe-expand-from-keymap-filter' work as expected."
         (funcall fn)
       (cl-loop for var in vars
                for saved in saved-values
-               do (set var saved)))))
+               do (unless (eq (symbol-value var) saved) ;Beware read-only vars!
+                    (set var saved))))))
 
 (defun yas-call-with-snippet-dirs (dirs fn)
   (let* ((default-directory (make-temp-file "yasnippet-fixture" t))
@@ -136,9 +137,6 @@ This lets `yas--maybe-expand-from-keymap-filter' work as expected."
 
 ;;; Older emacsen
 ;;;
-(unless (fboundp 'special-mode)
-  ;; FIXME: Why provide this default definition here?!?
-  (defalias 'special-mode 'fundamental))
 
 (unless (fboundp 'string-suffix-p)
   ;; introduced in Emacs 24.4
@@ -193,6 +191,20 @@ attention to case differences."
     (should (string= (yas--buffer-contents)
                      "bla from another BLA"))))
 
+(ert-deftest yas-mirror-many-fields ()
+  (with-temp-buffer
+    (yas-minor-mode 1)
+    (yas-expand-snippet "${1:brother} and ${2:brother} are${1:$(if (string= (yas-field-value 1) (yas-field-value 2)) \" \" \" not \")}the same word")
+    (should (string= (yas--buffer-contents)
+                     "brother and brother are the same word"))
+    (yas-mock-insert "bla")
+    (should (string= (yas--buffer-contents)
+                     "bla and brother are not the same word"))
+    (ert-simulate-command '(yas-next-field-or-maybe-expand))
+    (yas-mock-insert "bla")
+    (should (string= (yas--buffer-contents)
+                     "bla and bla are the same word"))))
+
 (ert-deftest mirror-with-transformation-and-autofill ()
   "Test interaction of autofill with mirror transforms"
   (let ((words "one two three four five")
@@ -274,10 +286,21 @@ attention to case differences."
     (ert-simulate-command '(yas-next-field-or-maybe-expand))
     (should (looking-at "testblable"))
     (ert-simulate-command '(yas-next-field-or-maybe-expand))
-    (ert-simulate-command '(yas-skip-and-clear-or-delete-char))
+    (ert-simulate-command '(yas-skip-and-clear-field))
     (should (looking-at "ble"))
     (should (null (yas-active-snippets)))))
 
+(ert-deftest delete-nested-simple-field-issue-824 ()
+  "Test deleting a field with a nested simple field in it."
+  (with-temp-buffer
+    (yas-minor-mode 1)
+    (yas-expand-snippet "${3:so-$4and}$0${2:-so}")
+    (ert-simulate-command '(yas-next-field-or-maybe-expand))
+    (should (looking-at "so-and-so"))
+    (ert-simulate-command '(yas-skip-and-clear-or-delete-char))
+    (should (looking-at "-so"))
+    (should (null (yas-active-snippets)))))
+
 (ert-deftest ignore-trailing-whitespace ()
   (should (equal
            (with-temp-buffer
@@ -297,40 +320,100 @@ attention to case differences."
 ;;     (should (string= (yas--buffer-contents)
 ;;                      "brother from another mother!"))))
 
-(ert-deftest undo-indentation ()
-  "Check undoing works when only line of snippet is indented."
-  (let ((yas-also-auto-indent-first-line t))
-    (yas-with-snippet-dirs
-     '((".emacs.d/snippets" ("emacs-lisp-mode" ("s" . "(setq $0)"))))
-     (with-temp-buffer
-       (emacs-lisp-mode)
-       (yas-reload-all)
-       (yas-minor-mode 1)
-       (insert "(let\n(while s")
+(ert-deftest undo-redo ()
+  "Check redoing of snippet undo."
+  (yas-with-snippet-dirs '((".emacs.d/snippets"
+                            ("emacs-lisp-mode" ("x" . "${1:one},and done"))))
+    (with-temp-buffer
+      (emacs-lisp-mode)
+      (yas-reload-all)
+      (yas-minor-mode 1)
+      (yas-expand-snippet "x$0")
+      (let ((pre-expand-string (buffer-string)))
+        (setq buffer-undo-list nil)
+        (ert-simulate-command '(yas-expand))
+        (push nil buffer-undo-list)
+        (ert-simulate-command '(yas-next-field)) ; $1 -> exit snippet.
+        (should (string-match-p "\\`one,and done" (buffer-string)))
+        (push nil buffer-undo-list)
+        (ert-simulate-command '(undo))  ; Revive snippet.
+        (ert-simulate-command '(undo))  ; Undo expansion.
+        (should (string= (buffer-string) pre-expand-string))
+        (ert-simulate-command '(move-end-of-line 1))
+        (push nil buffer-undo-list)
+        (ert-simulate-command '(undo))  ; Redo (re-expand snippet).
+        (should (string-match-p "\\`one,and done" (buffer-string)))))))
+
+(ert-deftest undo-revive-and-do-again ()
+  "Check undo-revived snippet is properly ended."
+  ;; See https://github.com/joaotavora/yasnippet/issues/1006.
+  (yas-with-snippet-dirs '((".emacs.d/snippets"
+                            ("emacs-lisp-mode" ("x" . "${1:one},and done"))))
+    (with-temp-buffer
+      (emacs-lisp-mode)
+      (yas-reload-all)
+      (yas-minor-mode 1)
+      (yas-expand-snippet "x$0")
+      (setq buffer-undo-list nil)
+      (ert-simulate-command '(yas-expand))
+      (push nil buffer-undo-list)
+      (ert-simulate-command '(yas-next-field)) ; $1 -> exit snippet.
+      (should (string-match-p "\\`one,and done" (buffer-string)))
+      (push nil buffer-undo-list)
+      (ert-simulate-command '(undo))    ; Revive snippet.
+      (yas-mock-insert "abc")
+      (ert-simulate-command '(yas-next-field)) ; $1 -> exit snippet again.
+      (should (string-match-p "\\`abc,and done" (buffer-string)))
+      ;; We should have exited snippet and cleaned up any overlays.
+      (should-not (cl-some (lambda (o) (overlay-get o 'yas--snippet))
+                           (overlays-in (point-min) (point-max)))))))
+
+
+(defun yas-test-expand-and-undo (mode snippet-entry initial-contents)
+  (yas-with-snippet-dirs
+   `((".emacs.d/snippets" (,(symbol-name mode) ,snippet-entry)))
+   (with-temp-buffer
+     (funcall mode)
+     (yas-reload-all)
+     (yas-minor-mode 1)
+     (yas-expand-snippet initial-contents)
+     (let ((pre-expand-string (buffer-string)))
        (setq buffer-undo-list ())
        (ert-simulate-command '(yas-expand))
        ;; Need undo barrier, I think command loop puts it normally.
        (push nil buffer-undo-list)
-       (should (string= (buffer-string) "(let\n    (while (setq )"))
        (ert-simulate-command '(undo))
-       (should (string= (buffer-string) "(let\n(while s"))))))
+       (should (string= (buffer-string) pre-expand-string))))))
 
-(ert-deftest undo-indentation-multiline ()
-  "Check undoing works when first line of multi-line snippet is indented."
-  (yas-with-snippet-dirs
-    '((".emacs.d/snippets" ("js-mode" ("if" . "if ($1) {\n\n}\n"))))
-    (with-temp-buffer
-      (js-mode)
-      (yas-reload-all)
-      (yas-minor-mode 1)
-      (insert "if\nabc = 123456789 + abcdef;")
-      (setq buffer-undo-list ())
-      (goto-char (point-min))
-      (search-forward "if")
-      (ert-simulate-command '(yas-expand))
-      (push nil buffer-undo-list)       ; See test above.
-      (ert-simulate-command '(undo))
-      (should (string= (buffer-string) "if\nabc = 123456789 + abcdef;")))))
+(ert-deftest undo-indentation-1 ()
+  "Check undoing works when only line of snippet is indented."
+  (let ((yas-also-auto-indent-first-line t))
+    (yas-test-expand-and-undo
+     'emacs-lisp-mode '("s" . "(setq $0)") "(let\n(while s$0")))
+
+(ert-deftest undo-indentation-2 ()
+  "Check undoing works when only line of snippet is indented."
+  (let ((yas-also-auto-indent-first-line t)
+        (indent-tabs-mode nil))
+    (yas-test-expand-and-undo
+     'emacs-lisp-mode '("t" . "; TODO") "t$0")))
+
+(ert-deftest undo-indentation-multiline-1 ()
+  "Check undoing works when 1st line of multi-line snippet is indented."
+  (let ((yas-also-auto-indent-first-line t)
+        (indent-tabs-mode nil))
+    (yas-test-expand-and-undo
+     'js-mode '("if" . "if ($1) {\n\n}\n")
+     "if$0\nabc = 123456789 + abcdef;")))
+
+
+(ert-deftest undo-indentation-multiline-2 ()
+  "Check undoing works when 2nd line of multi-line snippet is indented."
+  (let ((yas-also-auto-indent-first-line t)
+        (indent-tabs-mode nil))
+    (yas-test-expand-and-undo
+     'js-mode '("if" . "if (true) {\n${1:foo};\n}\n")
+     "if$0\nabc = 123456789 + abcdef;")))
 
 (ert-deftest dont-clear-on-partial-deletion-issue-515 ()
   "Ensure fields are not cleared when user doesn't really mean to."
@@ -384,6 +467,47 @@ end" (buffer-string)))
 end" (buffer-string)))
     (should (= 4 (current-column)))))
 
+(ert-deftest yas-also-indent-empty-lines ()
+  "Respect `yas-also-indent-empty-lines' setting."
+  (with-temp-buffer
+    (ruby-mode)
+    (yas-minor-mode 1)
+    (set (make-local-variable 'yas-indent-line) 'auto)
+    (set (make-local-variable 'yas-also-auto-indent-first-line) t)
+    (set (make-local-variable 'yas-also-indent-empty-lines) t)
+    (yas-expand-snippet "def foo\n\nend")
+    (should (string= "def foo\n  \nend" (buffer-string)))
+    ;; Test that it keeps working without setting
+    ;; `yas-also-auto-indent-first-line'.
+    (setq yas-also-auto-indent-first-line nil)
+    (erase-buffer)
+    (yas-expand-snippet "def foo\n\nend")
+    (should (string= "def foo\n  \nend" (buffer-string)))))
+
+(ert-deftest yas-indent-first-line ()
+  (with-temp-buffer
+    (ruby-mode)
+    (yas-minor-mode 1)
+    (set (make-local-variable 'yas-indent-line) 'auto)
+    (set (make-local-variable 'yas-also-auto-indent-first-line) nil)
+    (set (make-local-variable 'yas-also-indent-empty-lines) nil)
+    (yas-expand-snippet "def foo\n$0\nend\n")
+    ;; First (and only) line should not indent.
+    (yas-expand-snippet "#not indented")
+    (should (equal "def foo\n#not indented\nend\n" (buffer-string)))))
+
+(ert-deftest yas-indent-first-line-fixed ()
+  (with-temp-buffer
+    (ruby-mode)
+    (yas-minor-mode 1)
+    (set (make-local-variable 'yas-indent-line) 'fixed)
+    (set (make-local-variable 'yas-also-auto-indent-first-line) nil)
+    (set (make-local-variable 'yas-also-indent-empty-lines) nil)
+    (yas-expand-snippet "    def foo\n    $0\n    end\n")
+    ;; First (and only) line should not indent.
+    (yas-expand-snippet "#not more indented")
+    (should (equal "    def foo\n    #not more indented\n    end\n" (buffer-string)))))
+
 (ert-deftest indentation-markers ()
   "Test a snippet with indentation markers (`$<')."
   (with-temp-buffer
@@ -430,16 +554,19 @@ XXXXX   ------------------------"))))
     (yas-mock-insert "foo bar")
     (ert-simulate-command '(yas-next-field))
     (goto-char (point-min))
-    (let ((expected (with-temp-buffer
-                      (insert (format (concat "* Test foo bar\n"
-                                              "  " org-property-format "\n"
-                                              "  " org-property-format "\n"
-                                              "  " org-property-format)
-                                      ":PROPERTIES:" ""
-                                      ":ID:" "foo bar-after"
-                                      ":END:" ""))
-                      (delete-trailing-whitespace)
-                      (buffer-string))))
+    ;; The default value of `org-adapt-indentation' changed between Org-mode 9.4
+    ;; and 9.5, so force a specific value.
+    (let* ((org-adapt-indentation nil)
+           (expected (with-temp-buffer
+                       (insert (format (concat "* Test foo bar\n"
+                                               org-property-format "\n"
+                                               org-property-format "\n"
+                                               org-property-format)
+                                       ":PROPERTIES:" ""
+                                       ":ID:" "foo bar-after"
+                                       ":END:" ""))
+                       (delete-trailing-whitespace)
+                       (buffer-string))))
       ;; Some org-mode versions leave trailing whitespace, some don't.
       (delete-trailing-whitespace)
       (should (equal expected (buffer-string))))))
@@ -469,6 +596,22 @@ int foo()
   }
 }" (buffer-string)))))
 
+(ert-deftest indent-cc-mode-2 ()
+  "Handling of cc-mode's preprocessor indentation."
+  (with-temp-buffer
+    (c-mode)
+    (yas-minor-mode +1)
+    (yas-expand-snippet "\
+#ifndef `\"FOO\"`
+#define FOO
+#endif
+")
+    (should (string= "\
+#ifndef FOO
+#define FOO
+#endif
+" (buffer-substring-no-properties (point-min) (point-max))))))
+
 (ert-deftest indent-snippet-mode ()
   "Handling of snippet-mode indentation."
   ;; This is an interesting case because newlines match [[:space:]] in
@@ -491,6 +634,28 @@ int foo()
     ;; Assuming 2 space indent.
     (should (string= "def xxx\n  xxx\nend" (buffer-string)))))
 
+(defun yas-test-delete-and-insert-command (beg end new)
+  "Simulate a completion command (similar to company-mode)."
+  (interactive "r\ns")
+  ;; Simulate a completion command (like what company-mode does)
+  ;; which deletes the "xxx" and then replaces it with something
+  ;; else.
+  (delete-region beg end)
+  (insert new))
+
+(ert-deftest indent-mirrors-on-complex-update ()
+  "Don't get messed up by command that deletes and then inserts."
+  (with-temp-buffer
+    (ruby-mode)
+    (yas-minor-mode 1)
+    (yas-expand-snippet "def foo\n  ${1:slice} = append($1)\nend")
+    (yas-mock-insert "xxx")
+    (ert-simulate-command `(yas-test-delete-and-insert-command
+                            ,(- (point) 3) ,(point) ,"yyy"))
+    ;; Assuming 2 space indent.
+    (should (string= "def foo\n  yyy = append(yyy)\nend" (buffer-string)))))
+
+
 
 (ert-deftest snippet-with-multiline-mirrors-issue-665 ()
   "In issue 665, a multi-line mirror is attempted."
@@ -516,9 +681,37 @@ mapconcat #'(lambda (arg)
       (ert-simulate-command '(yas-next-field))
       (should (looking-at (concat "blo" expected))))))
 
+(defmacro yas-saving-variables (&rest body)
+  (declare (debug t))
+  `(yas-call-with-saving-variables #'(lambda () ,@body)))
+
+(ert-deftest auto-next-field ()
+  "Automatically exit a field after evaluating its transform."
+  (with-temp-buffer
+    (yas-saving-variables
+     (yas-with-snippet-dirs
+      `((".emacs.d/snippets"
+         ("ruby-mode" ("snip" . ,(concat "foo ${1:$$"
+                                         (prin1-to-string '(yas-auto-next
+                                                            (yas-choose-value
+                                                             "bar" "foo")))
+                                         "} ${2:$$"
+                                         (prin1-to-string '(yas-auto-next
+                                                            (yas-choose-value
+                                                             "too" "foo")))
+                                         "} baz ${3:quinn} quinn")))))
+      (yas-reload-all)
+      (ruby-mode)
+      (yas-minor-mode 1)
+      (set (make-local-variable 'yas-prompt-functions) `(yas-no-prompt))
+      (yas-mock-insert "snip")
+      (ert-simulate-command '(yas-expand))
+      (yas-mock-insert "quux")
+      (should (equal "foo bar too baz quux quinn" (buffer-string)))))))
+
 
 ;;; Snippet expansion and character escaping
-;;; Thanks to @zw963 (Billy) for the testing
+;; Thanks to @zw963 (Billy) for the testing
 ;;;
 (ert-deftest escape-dollar ()
   (with-temp-buffer
@@ -589,6 +782,57 @@ mapconcat #'(lambda (arg)
       (yas-expand-snippet "Look ma! ${1:`(yas-selected-text)`} OK?")
       (should (string= (yas--buffer-contents) "Look ma! He)}o world! OK?")))))
 
+(ert-deftest escaping-for-lsp-style-snippet-syntax ()
+  "See Github #979."
+  (should
+   (string= (with-temp-buffer
+              (yas-minor-mode 1)
+              (yas-expand-snippet
+               "Printf(${1:format string}, ${2:args ...interface{\\}})${0}")
+              (yas--buffer-contents))
+            (with-temp-buffer
+              (yas-minor-mode 1)
+              (yas-expand-snippet
+               "Printf(${1:format string}, ${2:args ...interface\\{\\}})${0}")
+              (yas--buffer-contents)))))
+
+(ert-deftest insert-snippet-with-backslashes-in-active-field ()
+  ;; This test case fails if `yas--inhibit-overlay-hooks' is not bound
+  ;; in `yas-expand-snippet' (see Github #844).
+  (with-temp-buffer
+    (yas-minor-mode 1)
+    (yas-expand-snippet "${1:$$(if (not yas-modified-p) \"a\")}")
+    (yas-expand-snippet "\\\\alpha")))
+
+(ert-deftest expand-with-unused-yas-selected-text ()
+  (with-temp-buffer
+    (yas-with-snippet-dirs
+      '((".emacs.d/snippets"
+         ("emacs-lisp-mode"
+          ("foo" . "expanded `yas-selected-text`foo"))))
+      (yas-reload-all)
+      (emacs-lisp-mode)
+      (yas-minor-mode +1)
+      (insert "foo")
+      (ert-simulate-command '(yas-expand))
+      (should (equal (buffer-string) "expanded foo")))))
+
+(ert-deftest yas-expand-command-snippet ()
+  (with-temp-buffer
+    (yas-with-snippet-dirs
+      '((".emacs.d/snippets"
+         ("emacs-lisp-mode"
+          ("foo" . "\
+# type: command
+# --
+\(insert \"expanded foo\")"))))
+      (yas-reload-all)
+      (emacs-lisp-mode)
+      (yas-minor-mode +1)
+      (insert "foo")
+      (ert-simulate-command '(yas-expand))
+      (should (equal (buffer-string) "expanded foo")))))
+
 (ert-deftest example-for-issue-271 ()
   (with-temp-buffer
     (yas-minor-mode 1)
@@ -599,8 +843,21 @@ mapconcat #'(lambda (arg)
       (yas-mock-insert "bbb")
       (should (string= (yas--buffer-contents) "if condition\naaa\nelse\nbbb\nend")))))
 
+(ert-deftest yas-no-memory-of-bad-snippet ()
+  "Expanding an incorrect snippet should not influence future expansions."
+  ;; See https://github.com/joaotavora/yasnippet/issues/800.
+  (with-temp-buffer
+    (yas-minor-mode 1)
+    (yas-expand-snippet "```foo\n\n```")
+    (erase-buffer) ; Bad snippet may leave wrong text.
+    ;; But expanding the corrected snippet should work fine.
+    (yas-expand-snippet "\\`\\`\\`foo\n\n\\`\\`\\`")
+    (should (equal (buffer-string) "```foo\n\n```"))))
+
 (defmacro yas--with-font-locked-temp-buffer (&rest body)
   "Like `with-temp-buffer', but ensure `font-lock-mode'."
+  ;; NOTE: Replace all uses of this with `font-lock-ensure' when we
+  ;; drop support for Emacs 24.
   (declare (indent 0) (debug t))
   (let ((temp-buffer (make-symbol "temp-buffer")))
     ;; NOTE: buffer name must not start with a space, otherwise
@@ -619,11 +876,9 @@ mapconcat #'(lambda (arg)
              (and (buffer-name ,temp-buffer)
                   (kill-buffer ,temp-buffer))))))))
 
-(defmacro yas-saving-variables (&rest body)
-  (declare (debug t))
-  `(yas-call-with-saving-variables #'(lambda () ,@body)))
-
 (ert-deftest example-for-issue-474 ()
+  ;; This issue only reproduces in Emacs 24.3, most likely due to some
+  ;; bug in the cc-mode included with that Emacs version.
   (yas--with-font-locked-temp-buffer
     (c-mode)
     (yas-minor-mode 1)
@@ -697,6 +952,18 @@ mapconcat #'(lambda (arg)
       (yas-mock-insert "baz")
       (should (string= (yas--buffer-contents) "foobaaarbazok")))))
 
+(ert-deftest yas-escaping-close-brace ()
+  "Close braces may be escaped with braces, reduction from eglot issue.
+See https://github.com/joaotavora/eglot/issues/336."
+  (with-temp-buffer
+    (yas-minor-mode +1)
+    ;; NOTE: put a period at the end to avoid the bug tested by
+    ;; `protection-overlay-no-cheating'.
+    (yas-expand-snippet "${1:one{\\}}, ${2:two{\\}}.")
+    (yas-next-field)
+    (yas-next-field)
+    (should (string= (buffer-string) "one{}, two{}."))))
+
 
 ;;; Misc tests
 ;;;
@@ -813,7 +1080,7 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
            ("'quote" . "OKquoteOK"))))
        (yas-reload-all)
        (emacs-lisp-mode)
-       (yas-minor-mode-on)
+       (yas-minor-mode +1)
        (let ((yas-key-syntaxes '("w" "w_")))
          (let ((yas--barbaz t))
            (yas-should-expand '(("foo-barbaz" . "foo-OKbarbazOK")
@@ -849,6 +1116,24 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
         (should (= (length (yas--snippet-fields (nth 0 snippets))) 2))
         (should (= (length (yas--snippet-fields (nth 1 snippets))) 1))))))
 
+(ert-deftest nested-snippet-expansion-depth-2 ()
+  (with-temp-buffer
+    (yas-with-snippet-dirs
+      '((".emacs.d/snippets"
+         ("text-mode"
+          ("nest" . "( $1"))))
+      (let ((yas-triggers-in-field t))
+        (yas-reload-all)
+        (text-mode)
+        (yas-minor-mode +1)
+        (dotimes (_ 3)
+          (yas-mock-insert "nest")
+          (ert-simulate-command '(yas-expand)))
+        (dotimes (_ 3)
+          (yas-mock-insert ")")
+          (ert-simulate-command '(yas-next-field-or-maybe-expand)))
+        ))))
+
 (ert-deftest nested-snippet-expansion-2 ()
   (let ((yas-triggers-in-field t))
     (yas-with-snippet-dirs
@@ -895,20 +1180,64 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
       (ert-simulate-command '(yas-next-field-or-maybe-expand))
       (should (string= (buffer-string) "\\sqrt[3]{\\sqrt[5]{2}}")))))
 
+(ert-deftest nested-snippet-expansion-4 ()
+  "See Github #959."
+  (let ((yas-triggers-in-field t))
+    (yas-with-snippet-dirs
+     '((".emacs.d/snippets"
+        ("text-mode"
+         ("ch" . "<-${1:ch}"))))
+     (yas-reload-all)
+     (text-mode)
+     (yas-minor-mode +1)
+     (yas-expand-snippet "ch$0\n")
+     (ert-simulate-command '(yas-expand))
+     (ert-simulate-command '(forward-char 2))
+     (ert-simulate-command '(yas-expand))
+     (yas-mock-insert "abc")
+     (ert-simulate-command '(yas-next-field-or-maybe-expand))
+     (yas-mock-insert "def")
+     (ert-simulate-command '(yas-next-field-or-maybe-expand))
+     (should (string= (buffer-string) "<-<-abcdef\n")))))
+
+(ert-deftest nested-snippet-expansion-5-nested-delete ()
+  "See Github #996."
+  (let ((yas-triggers-in-field t))
+    (yas-with-snippet-dirs
+     '((".emacs.d/snippets"
+        ("text-mode"
+         ("sel" . "${1:ch}")
+         ("ch" . "<-${1:ch}"))))
+     (yas-reload-all)
+     (text-mode)
+     (yas-minor-mode +1)
+     (insert "sel")
+     (ert-simulate-command '(yas-expand))
+     (ert-simulate-command '(forward-word 1))
+     (ert-simulate-command '(yas-expand))
+     (ert-simulate-command '(forward-word 1))
+     ;; The (cl-assert (memq pfield (yas--snippet-fields psnippet)))
+     ;; in `yas--on-field-overlay-modification' failed here.
+     (ert-simulate-command '(delete-backward-char 1))
+     (should (string= (buffer-string) "<-c\n")))))
+
 
 ;;; Loading
 ;;;
 
 (defmacro yas-with-overriden-buffer-list (&rest body)
   (declare (debug t))
+  ;; FIXME: This macro was added by commit 185c771dedea as part of the
+  ;; fix for https://github.com/joaotavora/yasnippet/issues/253,
+  ;; but I don't know why it was/is needed.
   (let ((saved-sym (make-symbol "yas--buffer-list")))
     `(let ((,saved-sym (symbol-function 'buffer-list)))
        (cl-letf (((symbol-function 'buffer-list)
-                  (lambda ()
+                  (lambda (&rest args)
                     (cl-remove-if (lambda (buf)
                                     (with-current-buffer buf
                                       (eq major-mode 'lisp-interaction-mode)))
-                                  (funcall ,saved-sym)))))
+                                  (apply ,saved-sym args)))))
          ,@body))))
 
 
@@ -936,11 +1265,28 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
   "Test `yas-lookup-snippet'."
   (yas-with-some-interesting-snippet-dirs
    (yas-reload-all 'no-jit)
-   (should (equal (yas-lookup-snippet "printf" 'c-mode) "printf($1);"))
-   (should (equal (yas-lookup-snippet "def" 'c-mode) "# define"))
+   (should (equal (yas--template-content (yas-lookup-snippet "printf" 'c-mode))
+                  "printf($1);"))
+   (should (equal (yas--template-content (yas-lookup-snippet "def" 'c-mode))
+                  "# define"))
    (should-not (yas-lookup-snippet "no such snippet" nil 'noerror))
    (should-not (yas-lookup-snippet "printf" 'emacs-lisp-mode 'noerror))))
 
+(ert-deftest yas-lookup-snippet-with-env ()
+  (with-temp-buffer
+    (yas-with-snippet-dirs
+      '((".emacs.d/snippets"
+         ("emacs-lisp-mode"
+          ("foo" . "\
+# expand-env: ((foo \"bar\"))
+# --
+`foo`"))))
+      (yas-reload-all)
+      (emacs-lisp-mode)
+      (yas-minor-mode +1)
+      (yas-expand-snippet (yas-lookup-snippet "foo"))
+      (should (equal (buffer-string) "bar")))))
+
 (ert-deftest basic-jit-loading ()
   "Test basic loading and expansion of snippets"
   (yas-with-some-interesting-snippet-dirs
@@ -968,13 +1314,15 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
    (with-temp-buffer
      (text-mode)
      (yas-minor-mode +1)
-     (should (equal (yas-lookup-snippet "one") "one"))
-     (should (eq (yas--key-binding "\C-c1") 'yas-expand-from-keymap))
+     (should (equal (yas--template-content (yas-lookup-snippet "one"))
+                    "one"))
+     (should (eq (yas--key-binding "\C-c1") #'yas-expand-from-keymap))
      (yas-define-snippets
       'text-mode '(("_1" "one!" "won" nil nil nil nil nil "uuid-1")))
      (should (null (yas-lookup-snippet "one" nil 'noerror)))
      (should (null (yas--key-binding "\C-c1")))
-     (should (equal (yas-lookup-snippet "won") "one!")))))
+     (should (equal (yas--template-content(yas-lookup-snippet "won"))
+                    "one!")))))
 
 (ert-deftest snippet-save ()
   "Make sure snippets can be saved correctly."
@@ -994,13 +1342,13 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
         (yas-minor-mode +1)
         (save-current-buffer
           (yas-new-snippet t)
-          (with-current-buffer "*new snippet*"
+          (with-current-buffer yas-new-snippet-buffer-name
             (snippet-mode)
             (insert "# name: foo\n# key: bar\n# --\nsnippet foo")
             (call-interactively 'yas-load-snippet-buffer-and-close)))
         (save-current-buffer
           (yas-new-snippet t)
-          (with-current-buffer "*new snippet*"
+          (with-current-buffer yas-new-snippet-buffer-name
             (snippet-mode)
             (insert "# name: bar\n# key: bar\n# --\nsnippet bar")
             (call-interactively 'yas-load-snippet-buffer-and-close)))
@@ -1036,19 +1384,20 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
      (yas-reload-all)
      (with-temp-buffer
        (let* ((major-mode 'c-mode)
-              (expected `(c-mode
+              (expected `(fundamental-mode
+                          c-mode
                           cc-mode
                           yet-another-c-mode
                           and-also-this-one
                           and-that-one
-                          ;; prog-mode doesn't exist in emacs 24.3
-                          ,@(if (fboundp 'prog-mode)
-                                '(prog-mode))
+                          prog-mode
+                          ,@(if (fboundp 'lisp-data-mode) ;Emacs≥28
+                                '(lisp-data-mode))
                           emacs-lisp-mode
                           lisp-interaction-mode))
               (observed (yas--modes-to-activate)))
          (should (equal major-mode (car observed)))
-         (should (equal (sort expected #'string<) (sort observed #'string<))))))))
+         (should-not (cl-set-exclusive-or expected observed)))))))
 
 (ert-deftest extra-modes-parenthood ()
   "Test activation of parents of `yas--extra-modes'."
@@ -1065,39 +1414,44 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
        (yas-activate-extra-mode 'and-that-one)
        (let* ((expected-first `(and-that-one
                                 yet-another-c-mode
-                                c-mode
-                                ,major-mode))
+                                c-mode))
               (expected-rest `(cc-mode
-                               ;; prog-mode doesn't exist in emacs 24.3
-                               ,@(if (fboundp 'prog-mode)
-                                     '(prog-mode))
+                               prog-mode
+                               ,@(if (fboundp 'lisp-data-mode) ;Emacs≥28
+                                     '(lisp-data-mode))
                                emacs-lisp-mode
                                and-also-this-one
                                lisp-interaction-mode))
-              (observed (yas--modes-to-activate)))
-         (should (equal expected-first
-                        (cl-subseq observed 0 (length expected-first))))
-         (should (equal (sort expected-rest #'string<)
-                        (sort (cl-subseq observed (length expected-first)) #'string<))))))))
+              (observed (remq 'fundamental-mode (yas--modes-to-activate))))
+         (should-not (cl-set-exclusive-or
+                      expected-first
+                      (cl-subseq observed 0 (length expected-first))))
+         (should-not (cl-set-exclusive-or
+                      expected-rest
+                      (cl-subseq observed (length expected-first)))))))))
 
-(defalias 'yas--phony-c-mode 'c-mode)
+(defalias 'yas--phony-c-mode #'c-mode)
 
 (ert-deftest issue-492-and-494 ()
+  "Aliases like `yas--phony-c-mode' should be considered as \"derived\"."
   (define-derived-mode yas--test-mode yas--phony-c-mode "Just a test mode")
   (yas-with-snippet-dirs '((".emacs.d/snippets"
                             ("yas--test-mode")))
-                         (yas-reload-all)
-                         (with-temp-buffer
-                           (let* ((major-mode 'yas--test-mode)
-                                  (expected `(c-mode
-                                              ,@(if (fboundp 'prog-mode)
-                                                    '(prog-mode))
-                                              yas--phony-c-mode
-                                              yas--test-mode))
-                                  (observed (yas--modes-to-activate)))
-                             (should (null (cl-set-exclusive-or expected observed)))
-                             (should (= (length expected)
-                                        (length observed)))))))
+    (yas-reload-all)
+    (with-temp-buffer
+      (let* ((major-mode 'yas--test-mode)
+             (expected `(fundamental-mode
+                         c-mode
+                         yas--phony-c-mode
+                         yas--test-mode))
+             ;; The set of mode depends on some external factors:
+             ;; `prog-mode': if cc-mode.el has been loaded.
+             ;; `cc-mode': if we added `cc-mode' as yas--parent of `c-mode'.
+             (observed (cl-set-difference (yas--modes-to-activate)
+                                          '(prog-mode cc-mode))))
+        (should-not (cl-set-exclusive-or expected observed))
+        (should (= (length expected)
+                   (length observed)))))))
 
 (define-derived-mode yas--test-mode c-mode "Just a test mode")
 (define-derived-mode yas--another-test-mode c-mode "Another test mode")
@@ -1140,6 +1494,34 @@ hello ${1:$(when (stringp yas-text) (funcall func yas-text))} foo${1:$$(concat \
                          ("def" . "# define")))
     (yas-should-not-expand '("sc" "dolist" "ert-deftest"))))
 
+
+;;; Unloading
+(ert-deftest yas-unload ()
+  "Test unloading and reloading."
+  (with-temp-buffer
+    (let ((status (call-process
+                   (concat invocation-directory invocation-name)
+                   nil '(t t) nil
+                   "-Q" "--batch" "-L" yas--loaddir "-l" "yasnippet"
+                   "--eval"
+                   (prin1-to-string
+                    '(condition-case err
+                         (progn
+                           (yas-minor-mode +1)
+                           (unload-feature 'yasnippet)
+                           ;; Unloading leaves `yas-minor-mode' bound,
+                           ;; harmless, though perhaps surprising.
+                           (when (bound-and-true-p yas-minor-mode)
+                             (error "`yas-minor-mode' still enabled"))
+                           (when (fboundp 'yas-minor-mode)
+                             (error "`yas-minor-mode' still fboundp"))
+                           (require 'yasnippet)
+                           (unless (fboundp 'yas-minor-mode)
+                             (error "Failed to reload")))
+                       (error (message "%S" (error-message-string err))
+                              (kill-emacs 1)))))))
+      (ert-info ((buffer-string)) (should (eq status 0))))))
+
 
 ;;; Menu
 ;;;
@@ -1273,14 +1655,14 @@ TODO: be meaner"
     (with-temp-buffer
       (yas-minor-mode -1)
       (insert "foo")
-      (should (not (eq (key-binding (yas--read-keybinding "<tab>")) 'yas-expand)))
+      (should-not (eq (key-binding (kbd "TAB")) #'yas-expand))
       (yas-minor-mode 1)
-      (should (eq (key-binding (yas--read-keybinding "<tab>")) 'yas-expand))
+      (should (eq (key-binding (kbd "TAB")) #'yas-expand))
       (yas-expand-snippet "$1 $2 $3")
-      (should (eq (key-binding [(tab)]) 'yas-next-field-or-maybe-expand))
-      (should (eq (key-binding (kbd "TAB")) 'yas-next-field-or-maybe-expand))
-      (should (eq (key-binding [(shift tab)]) 'yas-prev-field))
-      (should (eq (key-binding [backtab]) 'yas-prev-field))))))
+      ;; (should (eq (key-binding [tab]) #'yas-next-field-or-maybe-expand))
+      (should (eq (key-binding (kbd "TAB")) #'yas-next-field-or-maybe-expand))
+      (should (eq (key-binding [(shift tab)]) #'yas-prev-field))
+      (should (eq (key-binding [backtab]) #'yas-prev-field))))))
 
 (ert-deftest test-rebindings ()
   (let* ((yas-minor-mode-map (copy-keymap yas-minor-mode-map))
@@ -1290,14 +1672,14 @@ TODO: be meaner"
                            :test #'eq :key #'car))))
     (define-key yas-minor-mode-map [tab] nil)
     (define-key yas-minor-mode-map (kbd "TAB") nil)
-    (define-key yas-minor-mode-map (kbd "SPC") 'yas-expand)
+    (define-key yas-minor-mode-map (kbd "SPC") #'yas-expand)
     (with-temp-buffer
       (yas-minor-mode 1)
-      (should-not (eq (key-binding (kbd "TAB")) 'yas-expand))
-      (should (eq (key-binding (kbd "SPC")) 'yas-expand))
+      (should-not (eq (key-binding (kbd "TAB")) #'yas-expand))
+      (should (eq (key-binding (kbd "SPC")) #'yas-expand))
       (yas-reload-all)
-      (should-not (eq (key-binding (kbd "TAB")) 'yas-expand))
-      (should (eq (key-binding (kbd "SPC")) 'yas-expand)))))
+      (should-not (eq (key-binding (kbd "TAB")) #'yas-expand))
+      (should (eq (key-binding (kbd "SPC")) #'yas-expand)))))
 
 (ert-deftest test-yas-in-org ()
   (yas-saving-variables
@@ -1310,20 +1692,40 @@ TODO: be meaner"
       (org-mode)
       (yas-minor-mode 1)
       (insert "foo")
-      (should (eq (key-binding [(tab)]) 'yas-expand))
-      (should (eq (key-binding (kbd "TAB")) 'yas-expand))))))
+      ;; (should (eq (key-binding [tab]) #'yas-expand))
+      (should (eq (key-binding (kbd "TAB")) #'yas-expand))))))
 
-(ert-deftest yas-org-native-tab-in-source-block ()
+(ert-deftest yas-org-native-tab-in-source-block-text ()
+  "Test expansion of snippets in org source blocks."
+  ;; org 9+ no longer runs fontification for text-mode, so our hacks
+  ;; don't work.  Note that old ert doesn't have skipping, so we have
+  ;; to expect failure instead.  Starting with Org-mode 9.5 this seems
+  ;; to work again.
+  :expected-result (if (and (fboundp 'org-in-src-block-p)
+                            (or (version< (org-version) "9")
+                                (version<= "9.5" (org-version))))
+                       :passed :failed)
+  (let ((text-mode-hook #'yas-minor-mode))
+    (do-yas-org-native-tab-in-source-block "text")))
+
+(ert-deftest yas-org-native-tab-in-source-block-emacs-lisp ()
   "Test expansion of snippets in org source blocks."
   :expected-result (if (fboundp 'org-in-src-block-p)
                        :passed :failed)
+  (let ((emacs-lisp-mode-hook #'yas-minor-mode)
+        ;; This makes the test a bit less comprehensive, but it's
+        ;; needed to avoid bumping into Emacs Bug#35264.
+        (org-src-preserve-indentation t))
+    (do-yas-org-native-tab-in-source-block "emacs-lisp")))
+
+(defun do-yas-org-native-tab-in-source-block (mode)
   (yas-saving-variables
    (yas-with-snippet-dirs
-    '((".emacs.d/snippets"
-       ("text-mode"
+    `((".emacs.d/snippets"
+       (,(concat mode "-mode")
         ("T" . "${1:one} $1\n${2:two} $2\n<<$0>> done!"))))
-    (let ((text-mode-hook '(yas-minor-mode))
-          (org-src-tab-acts-natively t)
+    ;; Binding both text and prog mode hook should cover everything.
+    (let ((org-src-tab-acts-natively t)
           ;; Org 8.x requires this in order for
           ;; `org-src-tab-acts-natively' to have effect.
           (org-src-fontify-natively t))
@@ -1332,7 +1734,7 @@ TODO: be meaner"
       (yas--with-font-locked-temp-buffer
        (org-mode)
        (yas-minor-mode 1)
-       (insert "#+BEGIN_SRC text\nT\n#+END_SRC")
+       (insert "#+BEGIN_SRC " mode "\nT\n#+END_SRC")
        (if (fboundp 'font-lock-ensure)
            (font-lock-ensure)
          (jit-lock-fontify-now))
@@ -1348,9 +1750,9 @@ TODO: be meaner"
        ;; Check snippet expansion, ignore leading whitespace due to
        ;; `org-edit-src-content-indentation'.
        (should (looking-at "\
-[[:space:]]*one one
-[[:space:]]*two two
-[[:space:]]*<<>> done!")))))))
+\[[:space:]]*one one
+\[[:space:]]*two two
+\[[:space:]]*<<>> done!")))))))
 
 
 (ert-deftest test-yas-activate-extra-modes ()
@@ -1366,7 +1768,7 @@ add the snippets associated with the given mode."
            ("car" . "(car )"))))
        (yas-reload-all)
        (emacs-lisp-mode)
-       (yas-minor-mode-on)
+       (yas-minor-mode +1)
        (yas-activate-extra-mode 'markdown-mode)
        (should (eq 'markdown-mode (car yas--extra-modes)))
        (yas-should-expand '(("_" . "_Text_ ")))
@@ -1381,5 +1783,6 @@ add the snippets associated with the given mode."
 (provide 'yasnippet-tests)
 ;; Local Variables:
 ;; indent-tabs-mode: nil
+;; autoload-compute-prefixes: nil
 ;; End:
 ;;; yasnippet-tests.el ends here
